import copy
import os
# import sys
import json
import sqlalchemy
from docassemble.base.logger import logmessage
from docassemble.base.functions import server, this_thread
from docassemble.base.util import DAList, DAObjectPlusParameters
from docassemble.base.error import DAAttributeError, DAException
from alembic.config import Config
from alembic import command

__all__ = ['SQLObject', 'SQLObjectRelationship', 'SQLObjectList', 'SQLRelationshipList', 'StandardRelationshipList', 'alchemy_url', 'connect_args', 'upgrade_db']


class SQLObject:
    _required = []
    _uid = 'uid'
    _child_mapping = {}
    _parent_mapping = {}

    def sql_init(self):
        self._zombie = False
        db_entry = None
        if hasattr(self, 'id'):
            db_entry = self._session.query(self._model).filter(self._model.id == self.id).first()
            if db_entry is None:
                self._nascent = False
                self._zombie = True
                return
        if db_entry is None:
            db_entry = self.db_find_existing()
        if db_entry is not None:
            self.id = db_entry.id
            self._nascent = False
            db_values = {}
            for column in self._model.__dict__.keys():
                if column == 'id' or column.startswith('_'):
                    continue
                db_values[column] = getattr(db_entry, column)
                if db_values[column] is not None:
                    self.db_set(column, db_values[column])
                else:
                    try:
                        self.db_null(column)
                    except (DAAttributeError, AttributeError):
                        pass
                    except Exception as err:
                        logmessage("On SQLObject null, " + err.__class__.__name__ + ": " + str(err))
            self._orig = db_values
            self.db_cache()
            return
        self._nascent = True
        self._orig = {}

    def __getstate__(self):
        try:
            self.db_save()
        except Exception as err:
            logmessage("On SQLObject write, " + err.__class__.__name__ + ": " + str(err))
        dict_to_save = copy.copy(self.__dict__)
        if '_orig' in dict_to_save:
            del dict_to_save['_orig']
        return dict_to_save

    def __setstate__(self, pickle_dict):
        self.__dict__ = pickle_dict
        try:
            self.db_read()
        except Exception as err:
            logmessage("On SQLObject read, " + err.__class__.__name__ + ": " + str(err))

    @classmethod
    def filter(cls, instance_name, **kwargs):
        if 'dbcache' not in this_thread.misc:
            this_thread.misc['dbcache'] = {}
        listobj = DAList(instance_name, object_type=cls, auto_gather=False)
        filters = []
        for key, val in kwargs.items():
            if not hasattr(cls._model, key):
                raise DAException("filter: class " + cls.__name__ + " does not have column " + key)
            filters.append(getattr(cls._model, key) == val)
        for db_entry in list(cls._session.query(cls._model).filter(*filters).order_by(cls._model.id).all()):
            if cls._model.__name__ in this_thread.misc['dbcache'] and db_entry.id in this_thread.misc['dbcache'][cls._model.__name__]:
                listobj.append(this_thread.misc['dbcache'][cls._model.__name__][db_entry.id])
            else:
                obj = listobj.appendObject()
                obj.id = db_entry.id
                db_values = {}
                for column in cls._model.__dict__.keys():
                    if column == 'id' or column.startswith('_'):
                        continue
                    db_values[column] = getattr(db_entry, column)
                    if db_values[column] is not None:
                        obj.db_set(column, db_values[column])
                obj._orig = db_values
                obj.db_cache()
        listobj.gathered = True
        return listobj

    @classmethod
    def any(cls):
        db_entry = cls._session.query(cls._model).first()
        if db_entry is None:
            return False
        return True

    @classmethod
    def all(cls, instance_name=None):
        if 'dbcache' not in this_thread.misc:
            this_thread.misc['dbcache'] = {}
        if instance_name:
            listobj = DAList(instance_name, object_type=cls)
        else:
            listobj = DAList(object_type=cls)
            listobj.set_random_instance_name()
        for db_entry in list(cls._session.query(cls._model).order_by(cls._model.id).all()):
            if cls._model.__name__ in this_thread.misc['dbcache'] and db_entry.id in this_thread.misc['dbcache'][cls._model.__name__]:
                listobj.append(this_thread.misc['dbcache'][cls._model.__name__][db_entry.id])
            else:
                obj = listobj.appendObject()
                obj.id = db_entry.id
                db_values = {}
                for column in cls._model.__dict__.keys():
                    if column == 'id' or column.startswith('_'):
                        continue
                    db_values[column] = getattr(db_entry, column)
                    if db_values[column] is not None:
                        obj.db_set(column, db_values[column])
                obj._orig = db_values
                obj.db_cache()
        listobj.gathered = True
        return listobj

    @classmethod
    def by_id(cls, the_id, instance_name=None):
        if 'dbcache' not in this_thread.misc:
            this_thread.misc['dbcache'] = {}
        if cls._model.__name__ in this_thread.misc['dbcache'] and the_id in this_thread.misc['dbcache'][cls._model.__name__]:
            if instance_name is None:
                return this_thread.misc['dbcache'][cls._model.__name__][the_id]
            obj = this_thread.misc['dbcache'][cls._model.__name__][the_id]
            obj.fix_instance_name(obj.instanceName, instance_name)
        if instance_name is None:
            obj = cls(id=the_id)
            obj.set_random_instance_name()
        else:
            obj = cls(instance_name, id=the_id)
        return obj

    @classmethod
    def by_uid(cls, uid, instance_name=None):
        if cls._uid not in cls._model.__dict__:
            return None
        db_entry = cls._session.query(cls._model).filter(getattr(cls._model, cls._uid) == uid).first()
        if db_entry is not None:
            return cls.by_id(db_entry.id)
        if instance_name is None:
            obj = cls()
        else:
            obj = cls(instance_name)
        obj.db_set(cls._uid, uid)
        if instance_name is None:
            obj.set_random_instance_name()
        obj.ready()
        return obj

    @classmethod
    def delete_by_id(cls, the_id):
        cls._session.query(cls._model).filter(cls._model.id == the_id).delete()
        cls._session.commit()
        if 'dbcache' in this_thread.misc and cls._model.__name__ in this_thread.misc['dbcache'] and the_id in this_thread.misc['dbcache'][cls._model.__name__]:
            this_thread.misc['dbcache'][cls._model.__name__][the_id]._zombie = True

    @classmethod
    def delete_by_uid(cls, uid):
        if cls._uid not in cls._model.__dict__:
            return
        db_entry = cls._session.query(cls._model).filter(getattr(cls._model, cls._uid) == uid).first()
        if db_entry is not None:
            the_id = db_entry.id
            cls._session.query(cls._model).filter(getattr(cls._model, cls._uid) == uid).delete()
            cls._session.commit()
            if 'dbcache' in this_thread.misc and cls._model.__name__ in this_thread.misc['dbcache'] and the_id in this_thread.misc['dbcache'][cls._model.__name__]:
                this_thread.misc['dbcache'][cls._model.__name__][the_id]._zombie = True

    @classmethod
    def id_exists(cls, the_id):
        db_entry = cls._session.query(cls._model).filter(cls._model.id == the_id).first()
        if db_entry is None:
            return False
        return True

    @classmethod
    def uid_exists(cls, uid):
        if cls._uid not in cls._model.__dict__:
            return False
        db_entry = cls._session.query(cls._model).filter(getattr(cls._model, cls._uid) == uid).first()
        if db_entry is None:
            return False
        return True

    def db_from_cache(self, the_id):
        if 'dbcache' not in this_thread.misc:
            this_thread.misc['dbcache'] = {}
        if self._model.__name__ not in this_thread.misc['dbcache']:
            this_thread.misc['dbcache'][self._model.__name__] = {}
        if the_id in this_thread.misc['dbcache'][self._model.__name__]:
            return this_thread.misc['dbcache'][self._model.__name__][the_id]
        return None

    def db_cache(self):
        if 'dbcache' not in this_thread.misc:
            this_thread.misc['dbcache'] = {}
        if self._model.__name__ not in this_thread.misc['dbcache']:
            this_thread.misc['dbcache'][self._model.__name__] = {}
        if hasattr(self, 'id'):
            this_thread.misc['dbcache'][self._model.__name__][self.id] = self

    def __del__(self):
        if hasattr(self, 'id') and 'dbcache' in this_thread.misc and self._model.__name__ in this_thread.misc['dbcache'] and self.id in this_thread.misc['dbcache'][self._model.__name__]:
            del this_thread.misc['dbcache'][self._model.__name__][self.id]

    def db_delete(self):
        self.db_read()
        if self._nascent or self._zombie:
            return
        if hasattr(self, 'id'):
            self.delete_by_id(self.id)
            self._zombie = True

    def db_save(self):
        if self._zombie:
            return
        db_values = {}
        required_ok = True
        for column in self._model.__dict__.keys():
            if column == 'id' or column.startswith('_'):
                continue
            try:
                db_values[column] = self.db_get(column)  # pylint: disable=assignment-from-no-return
            except:
                if column == self._uid:
                    return
                if column in self._required:
                    if not self._nascent:
                        return
                    required_ok = False
        if self._nascent:
            try:
                db_entry = self.db_find_existing()
                assert db_entry is not None
                for key, val in db_values.items():
                    setattr(db_entry, key, val)
                new_db_values = {}
                for column in self._model.__dict__.keys():
                    if column not in db_values:
                        new_db_values[column] = getattr(db_entry, column)
                        if new_db_values[column] is not None:
                            self.db_set(column, new_db_values[column])
                        else:
                            try:
                                self.db_null(column)
                            except (DAAttributeError, AttributeError):
                                pass
                            except Exception as err:
                                logmessage("On SQLObject null, " + err.__class__.__name__ + ": " + str(err))
            except Exception as err:
                logmessage(err.__class__.__name__ + ": " + str(err))
                if not required_ok:
                    return
                db_entry = self._model(**db_values)
                self._session.add(db_entry)
            self._session.commit()
            self.id = db_entry.id
            self._nascent = False
        else:
            if self._orig == db_values:
                return
            db_entry = self._session.query(self._model).filter(self._model.id == self.id).first()
            if db_entry is None:
                self._zombie = True
                return
            for key, val in db_values.items():
                setattr(db_entry, key, val)
            null_keys = []
            for key, val in self._orig.items():
                if key not in db_values:
                    setattr(db_entry, key, None)
                    null_keys.append(key)
            self._session.commit()
            self._orig = db_values
            for key in null_keys:
                self._orig[key] = None

    def save_if_nascent(self):
        if self._nascent:
            self.db_save()

    def ready(self):
        self.save_if_nascent()
        return not (self._nascent or self._zombie)

    def db_read(self):
        if self._zombie:
            self.db_cache()
            return
        if hasattr(self, 'id'):
            db_entry = self._session.query(self._model).filter(self._model.id == self.id).first()
        else:
            try:
                db_entry = self.db_find_existing()
            except Exception as err:
                logmessage("On SQLObject read, " + err.__class__.__name__ + ": " + str(err))
                db_entry = None
            if db_entry is None and self._nascent:
                self.db_cache()
                return
            if db_entry is not None:
                self.id = db_entry.id
        if db_entry is None:
            self._zombie = True
            self.db_cache()
            return
        if self._nascent is True:
            self._nascent = False
        db_values = {}
        for column in self._model.__dict__.keys():
            if column == 'id' or column.startswith('_'):
                continue
            db_values[column] = getattr(db_entry, column)
            if db_values[column] is not None:
                self.db_set(column, db_values[column])
            else:
                try:
                    self.db_null(column)
                except (DAAttributeError, AttributeError):
                    pass
                except Exception as err:
                    logmessage("On SQLObject null, " + err.__class__.__name__ + ": " + str(err))
        self._orig = db_values
        self.db_cache()

    def db_find_existing(self):
        if self._uid not in self._model.__dict__:
            return None
        try:
            the_uid = self.db_get(self._uid)  # pylint: disable=assignment-from-no-return
        except:
            return None
        return self._session.query(self._model).filter(getattr(self._model, self._uid) == the_uid).first()

    def db_get(self, column):  # pylint: disable=unused-argument
        pass

    def db_set(self, column, value):  # pylint: disable=unused-argument
        pass

    def db_null(self, column):  # pylint: disable=unused-argument
        pass

    def has_child(self, rel_name, rel):
        if not (self.ready() and rel.ready()):
            raise DAException("has_child: cannot retrieve data")
        info = self._child_mapping[rel_name]
        model = info['relationship_class']._model
        db_entry = self._session.query(model).filter(model.getattr(info['parent_column']) == self.id, model.getattr(info['child_column']) == rel.id).first()
        if db_entry is None:
            return False
        return True

    def add_child(self, rel_name, child):
        if not self.has_child(rel_name, child):
            info = self._child_mapping[rel_name]
            arg_dict = {}
            arg_dict[info['parent_column']] = self.id
            arg_dict[info['child_column']] = child.id
            db_entry = info['relationship_class']._model(**arg_dict)
            self._session.add(db_entry)
            self._session.commit()

    def get_child(self, rel_name, instance_name=None):
        if not self.ready():
            raise DAException("get_child: cannot retrieve data")
        info = self._child_mapping[rel_name]
        model = info['relationship_class']._model
        if instance_name:
            results = DAList(instance_name, object_type=info['parent_class'])
        else:
            results = []
        indexno = 0
        if instance_name is None:
            for db_entry in list(self._session.query(model).filter(model.getattr(info['parent_column']) == self.id).all()):
                results.append(info['child_class'].by_id(db_entry.getattr(info['child_column'])))
        else:
            for db_entry in list(self._session.query(model).filter(model.getattr(info['parent_column']) == self.id).all()):
                results.append(info['child_class'].by_id(db_entry.getattr(info['child_column'])), instance_name=instance_name + '[' + str(indexno) + ']')
                indexno += 1
        return results

    def del_child(self, rel_name, child):
        if not (self.ready() and child.ready()):
            raise DAException("del_child: cannot retrieve data")
        info = self._child_mapping[rel_name]
        model = info['relationship_class']._model
        self._session.query(model).filter(model.getattr(info['parent_column']) == self.id, model.getattr(info['child_column']) == child.id).delete()
        self._session.commit()

    def has_parent(self, rel_name, rel):
        if not (self.ready() and rel.ready()):
            raise DAException("has_parent: cannot retrieve data")
        info = self._parent_mapping[rel_name]
        model = info['relationship_class']._model
        db_entry = self._session.query(model).filter(model.getattr(info['child_column']) == self.id, model.getattr(info['parent_column']) == rel.id).first()
        if db_entry is None:
            return False
        return True

    def add_parent(self, rel_name, parent):
        if not self.has_parent(rel_name, parent):
            info = self._parent_mapping[rel_name]
            arg_dict = {}
            arg_dict[info['child_column']] = self.id
            arg_dict[info['parent_column']] = parent.id
            db_entry = info['relationship_class']._model(**arg_dict)
            self._session.add(db_entry)
            self._session.commit()

    def get_parent(self, rel_name, instance_name=None):
        if not self.ready():
            raise DAException("get_parent: cannot retrieve data")
        info = self._parent_mapping[rel_name]
        model = info['relationship_class']._model
        if instance_name:
            results = DAList(instance_name, object_type=info['parent_class'])
        else:
            results = []
        indexno = 0
        if instance_name is None:
            for db_entry in list(self._session.query(model).filter(model.getattr(info['child_column']) == self.id).all()):
                results.append(info['parent_class'].by_id(db_entry.getattr(info['parent_column'])))
        else:
            for db_entry in list(self._session.query(model).filter(model.getattr(info['child_column']) == self.id).all()):
                results.append(info['parent_class'].by_id(db_entry.getattr(info['parent_column'])), instance_name=instance_name + '[' + str(indexno) + ']')
                indexno += 1
        return results

    def del_parent(self, rel_name, parent):
        if not (self.ready() and parent.ready()):
            raise DAException("del_parent: cannot retrieve data")
        info = self._parent_mapping[rel_name]
        model = info['relationship_class']._model
        self._session.query(model).filter(model.getattr(info['child_column']) == self.id, model.getattr(info['parent_column']) == parent.id).delete()
        self._session.commit()


def alchemy_url(db_config):
    """Returns a URL representing a database connection."""
    return server.alchemy_url(db_config)


def connect_args(db_config):
    """Returns PostgreSQL arguments for connecting via SSL."""
    return server.connect_args(db_config)


def upgrade_db(url, py_file, engine, version_table=None, name=None, conn_args=None):
    if name is None:
        name = 'alembic'
    if version_table is None:
        version_table = 'alembic_version'
    elif version_table == 'auto':
        version_table = 'alembic_version_' + os.path.abspath(py_file).split(os.sep)[-2]
    if not isinstance(conn_args, dict):
        conn_args = {}
    packagedir = os.path.dirname(os.path.abspath(py_file))
    alembic_path = os.path.join(packagedir, name)
    if not os.path.isdir(alembic_path):
        # logmessage(name + " directory not found in package directory " + packagedir)
        return
    ini_file = os.path.join(packagedir, name + '.ini')
    if not os.path.isfile(ini_file):
        logmessage(name + ".ini file not found at " + ini_file)
        return
    versions_path = os.path.join(alembic_path, 'versions')
    if not os.path.isdir(versions_path):
        os.makedirs(versions_path, exist_ok=True)
    alembic_cfg = Config(ini_file)
    alembic_cfg.set_main_option("sqlalchemy.url", url)
    alembic_cfg.set_main_option("connect_args", json.dumps(conn_args))
    alembic_cfg.set_main_option("script_location", alembic_path)
    if not sqlalchemy.inspect(engine).has_table(version_table):
        command.stamp(alembic_cfg, "head")
    command.upgrade(alembic_cfg, "head")


class SQLObjectRelationship(SQLObject):

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._required = [cls._parent[2], cls._child[2]]
        if cls.__name__ not in cls._parent[0]._child_mapping:
            cls._parent[0]._child_mapping[cls.__name__] = {}
        cls._parent[0]._child_mapping[cls.__name__]['relationship_class'] = cls
        cls._parent[0]._child_mapping[cls.__name__]['child_class'] = cls._child[0]
        cls._parent[0]._child_mapping[cls.__name__]['parent_column'] = cls._parent[2]
        cls._parent[0]._child_mapping[cls.__name__]['child_column'] = cls._child[2]
        if cls.__name__ not in cls._child[0]._parent_mapping:
            cls._child[0]._parent_mapping[cls.__name__] = {}
        cls._child[0]._parent_mapping[cls.__name__]['relationship_class'] = cls
        cls._child[0]._parent_mapping[cls.__name__]['parent_class'] = cls._parent[0]
        cls._child[0]._parent_mapping[cls.__name__]['child_column'] = cls._child[2]
        cls._child[0]._parent_mapping[cls.__name__]['parent_column'] = cls._parent[2]

    @classmethod
    def filter_by_parent(cls, instance_name, parent):
        if not parent.ready():
            raise DAException("filter_by_parent: cannot retrieve data")
        filters = {}
        filters[cls._parent[2]] = parent.id
        return cls.filter(instance_name, **filters)

    @classmethod
    def filter_by_child(cls, instance_name, child):
        if not child.ready():
            raise DAException("filter_by_child: cannot retrieve data")
        filters = {}
        filters[cls._child[2]] = child.id
        return cls.filter(instance_name, **filters)

    @classmethod
    def filter_by_parent_child(cls, instance_name, parent, child):
        if not (child.ready() and parent.ready()):
            raise DAException("filter_by_parent_child: cannot retrieve data")
        filters = {}
        filters[cls._parent[2]] = parent.id
        filters[cls._child[2]] = child.id
        return cls.filter(instance_name, **filters)

    def rel_init(self, *pargs, **kwargs):  # pylint: disable=unused-argument
        self.sql_init()
        if not hasattr(self, self._parent[1]):
            self.initializeAttribute(self._parent[1], self._parent[0])
        if not hasattr(self, self._child[1]):
            self.initializeAttribute(self._child[1], self._child[0])

    def db_get(self, column):
        if column == self._parent[2]:
            return getattr(getattr(self, self._parent[1]), 'id')
        if column == self._child[2]:
            return getattr(getattr(self, self._child[1]), 'id')
        raise DAException("Invalid column " + column)

    def db_set(self, column, value):
        if column == self._parent[2]:
            self.reInitializeAttribute(self._parent[1], self._parent[0].using(id=value))
        elif column == self._child[2]:
            self.reInitializeAttribute(self._child[1], self._child[0].using(id=value))

    def db_find_existing(self):
        return self._session.query(self._model).filter(getattr(self._model, self._parent[2]) == getattr(getattr(self, self._parent[1]), 'id'), getattr(self._model, self._child[2]) == getattr(getattr(self, self._child[1]), 'id')).first()


class SQLObjectList(DAList):

    def init(self, *pargs, **kwargs):
        all_arg = kwargs.get('all', False)
        if 'all' in kwargs:
            del kwargs['all']
        super().init(*pargs, **kwargs)
        if all_arg:
            self.elements = self.object_type.all(self.instanceName)


class SQLRelationshipList(DAList):

    def init(self, *pargs, **kwargs):
        parent_arg = kwargs.get('parent', None)
        child_arg = kwargs.get('child', None)
        if 'object_type' in kwargs:
            params = {}
            if 'parent' in kwargs:
                params[kwargs['object_type']._parent[1]] = kwargs['parent']
            if 'child' in kwargs:
                params[kwargs['object_type']._child[1]] = kwargs['child']
            if isinstance(kwargs['object_type'], DAObjectPlusParameters):
                for key, val in params:
                    if key not in kwargs['object_type'].parameters:
                        kwargs['object_type'].parameters[key] = val
            else:
                kwargs['object_type'] = kwargs['object_type'].using(**params)
        if 'parent' in kwargs:
            del kwargs['parent']
        if 'child' in kwargs:
            del kwargs['child']
        super().init(*pargs, **kwargs)
        if parent_arg and not child_arg:
            self.elements = self.object_type.filter_by_parent(self.instanceName, parent_arg)
        elif child_arg and not parent_arg:
            self.elements = self.object_type.filter_by_child(self.instanceName, child_arg)
        elif child_arg and parent_arg:
            self.elements = self.object_type.filter_by_parent_child(self.instanceName, parent_arg, child_arg)

    def hook_on_remove(self, item, *pargs, **kwargs):
        item.db_delete()

    def reset_gathered(self, recursive=False, only_if_empty=False, mark_incomplete=False):
        return


class StandardRelationshipList(SQLRelationshipList):

    def init(self, *pargs, **kwargs):
        super().init(*pargs, **kwargs)
        self.there_is_another = False
        self.gathered = True
        self.complete_attribute = 'complete'



import string
import random
import sys
import math

__all__ = ['random_string', 'random_alphanumeric']

r = random.SystemRandom()


def random_bytes(length):
    return bytearray(ord(r.choice(string.ascii_letters)) for i in range(length))


def random_lower_string(length):
    return ''.join(r.choice(string.ascii_lowercase) for i in range(length))


def random_string(length):
    return ''.join(r.choice(string.ascii_letters) for i in range(length))


def random_alphanumeric(length):
    return ''.join(r.choice(string.ascii_uppercase + string.ascii_lowercase + string.digits) for x in range(length))


def random_digits(num):
    the_string = ("%0" + str(num) + "d") % int(random.random()*math.pow(10, int(num)))
    return the_string[0:int(num)]

if __name__ == "__main__":
    sys.stdout.write(random_string(32))



import ast
import re

# import sys

fix_assign = re.compile(r'\.(\[[^\]]*\])')
valid_variable_match = re.compile(r'^[^\d][A-Za-z0-9\_]*$')


class myextract(ast.NodeVisitor):

    def __init__(self):
        self.stack = []
        self.in_subscript = 0
        self.in_params = 0
        self.seen_name = False
        self.seen_complexity = False

    def visit_Name(self, node):
        if not (self.in_subscript > 0 and self.seen_name is True):
            self.stack.append(node.id)
            if self.in_subscript > 0:
                self.seen_name = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Call(self, node):
        self.seen_complexity = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Attribute(self, node):
        self.stack.append(node.attr)
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Subscript(self, node):
        if isinstance(node.slice, ast.Name):
            self.stack.append('[' + str(node.slice.id) + ']')
            self.in_subscript += 1
            self.seen_name = False
        elif isinstance(node.slice, ast.Constant):
            self.stack.append('[' + repr(node.slice.value) + ']')
            self.in_subscript += 1
            self.seen_name = False
        elif isinstance(node.slice, ast.Index):
            if isinstance(node.slice.value, ast.Name):
                self.stack.append('[' + str(node.slice.value.id) + ']')
                self.in_subscript += 1
                self.seen_name = False
            elif isinstance(node.slice.value, ast.Constant):
                self.stack.append('[' + repr(node.slice.value.value) + ']')
                self.in_subscript += 1
                self.seen_name = False
            elif hasattr(node.slice.value, 'n'):
                self.stack.append('[' + repr(node.slice.value.n) + ']')
                self.in_subscript += 1
                self.seen_name = False
            elif hasattr(node.slice.value, 's'):
                self.stack.append('[' + repr(str(node.slice.value.s)) + ']')
                self.in_subscript += 1
                self.seen_name = False
            else:
                self.seen_complexity = 1
        else:
            self.seen_complexity = 1
        ast.NodeVisitor.generic_visit(self, node)
        if hasattr(node.slice, 'slice') and (hasattr(node.slice.value, 'id') or hasattr(node.slice.value, 'n') or hasattr(node.slice.value, 'value')):
            self.in_subscript -= 1


class myvisitnode(ast.NodeVisitor):

    def __init__(self):
        self.names = {}
        self.targets = {}
        self.depth = 0
        self.calls = set()

    def generic_visit(self, node):
        # logmessage(' ' * self.depth + type(node).__name__)
        self.depth += 1
        ast.NodeVisitor.generic_visit(self, node)
        self.depth -= 1

    def visit_Call(self, node):
        self.calls.add(node.func)
        if hasattr(node.func, 'id') and node.func.id in ['showif', 'showifdef', 'value', 'defined'] and len(node.args) and node.args[0].__class__.__name__ == 'Constant' and hasattr(node.args[0], 'value') and re.search(r'^[^\d]', node.args[0].value) and not re.search(r'[^A-Z_a-z0-9\.\"\'\[\] ]', node.args[0].value):
            self.names[node.args[0].value] = 1
        if hasattr(node.func, 'id') and node.func.id in ['define'] and len(node.args) and node.args[0].__class__.__name__ == 'Constant' and hasattr(node.args[0], 'value') and re.search(r'^[^\d]', node.args[0].value) and not re.search(r'[^A-Z_a-z0-9\.\"\'\[\] ]', node.args[0].value):
            self.targets[node.args[0].value] = 1
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Subscript(self, node):
        if node not in self.calls:
            crawler = myextract()
            crawler.visit(node)
            if not crawler.seen_complexity:
                self.names[fix_assign.sub(r'\1', (".".join(reversed(crawler.stack))))] = 1
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Attribute(self, node):
        if node not in self.calls:
            crawler = myextract()
            crawler.visit(node)
            if not crawler.seen_complexity:
                self.names[fix_assign.sub(r'\1', (".".join(reversed(crawler.stack))))] = 1
        ast.NodeVisitor.generic_visit(self, node)

    def visit_ExceptHandler(self, node):
        if node.name is not None and hasattr(node.name, 'id') and node.name.id is not None:
            self.targets[node.name.id] = 1
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Assign(self, node):
        for key, val in ast.iter_fields(node):
            if key == 'targets':
                for subnode in val:
                    if isinstance(subnode, ast.Tuple):
                        for subsubnode in subnode.elts:
                            crawler = myextract()
                            crawler.visit(subsubnode)
                            self.targets[fix_assign.sub(r'\1', ".".join(reversed(crawler.stack)))] = 1
                    else:
                        crawler = myextract()
                        crawler.visit(subnode)
                        self.targets[fix_assign.sub(r'\1', ".".join(reversed(crawler.stack)))] = 1
        self.depth += 1
        # ast.NodeVisitor.generic_visit(self, node)
        self.generic_visit(node)
        self.depth -= 1

    def visit_AugAssign(self, node):
        for key, val in ast.iter_fields(node):
            if key == 'target':
                crawler = myextract()
                crawler.visit(val)
                self.targets[fix_assign.sub(r'\1', ".".join(reversed(crawler.stack)))] = 1
        self.depth += 1
        # ast.NodeVisitor.generic_visit(self, node)
        self.generic_visit(node)
        self.depth -= 1

    def visit_AnnAssign(self, node):
        for key, val in ast.iter_fields(node):
            if key == 'target':
                crawler = myextract()
                crawler.visit(val)
                self.targets[fix_assign.sub(r'\1', ".".join(reversed(crawler.stack)))] = 1
        self.depth += 1
        # ast.NodeVisitor.generic_visit(self, node)
        self.generic_visit(node)
        self.depth -= 1

    def visit_FunctionDef(self, node):
        if hasattr(node, 'name'):
            self.targets[node.name] = 1

    def visit_Import(self, node):
        for alias in node.names:
            if alias.asname is None:
                the_name = alias.name
            else:
                the_name = alias.asname
            while re.search(r'\.', the_name):
                self.targets[the_name] = 1
                the_name = re.sub(r'\.[^\.]+$', '', the_name)
            self.targets[the_name] = 1

    def visit_ImportFrom(self, node):
        for alias in node.names:
            if alias.asname is None:
                the_name = alias.name
            else:
                the_name = alias.asname
            while re.search(r'\.', the_name):
                self.targets[the_name] = 1
                the_name = re.sub(r'\.[^\.]+$', '', the_name)
            self.targets[the_name] = 1

    def visit_GeneratorExp(self, node):
        for comp in node.generators:
            if isinstance(comp.target, ast.Name):
                self.targets[comp.target.id] = 1
            elif isinstance(comp.target, ast.Tuple):
                for subtarget in comp.target.elts:
                    if isinstance(subtarget, ast.Name):
                        self.targets[subtarget.id] = 1
        self.generic_visit(node)

    def visit_Lambda(self, node):
        for arg in node.args.posonlyargs:
            self.targets[arg.arg] = 1
        for arg in node.args.args:
            self.targets[arg.arg] = 1
        for arg in node.args.kwonlyargs:
            self.targets[arg.arg] = 1
        self.generic_visit(node)

    def visit_ListComp(self, node):
        for comp in node.generators:
            if isinstance(comp.target, ast.Name):
                self.targets[comp.target.id] = 1
            elif isinstance(comp.target, ast.Tuple):
                for subtarget in comp.target.elts:
                    if isinstance(subtarget, ast.Name):
                        self.targets[subtarget.id] = 1
        self.generic_visit(node)

    def visit_DictComp(self, node):
        for comp in node.generators:
            if isinstance(comp.target, ast.Name):
                self.targets[comp.target.id] = 1
            elif isinstance(comp.target, ast.Tuple):
                for subtarget in comp.target.elts:
                    if isinstance(subtarget, ast.Name):
                        self.targets[subtarget.id] = 1
        self.generic_visit(node)

    def visit_SetComp(self, node):
        for comp in node.generators:
            if isinstance(comp.target, ast.Name):
                self.targets[comp.target.id] = 1
            elif isinstance(comp.target, ast.Tuple):
                for subtarget in comp.target.elts:
                    if isinstance(subtarget, ast.Name):
                        self.targets[subtarget.id] = 1
        self.generic_visit(node)

    def visit_For(self, node):
        if isinstance(node.target, ast.Name):
            self.targets[node.target.id] = 1
        elif isinstance(node.target, ast.Tuple):
            for subtarget in node.target.elts:
                if isinstance(subtarget, ast.Name):
                    self.targets[subtarget.id] = 1
        self.generic_visit(node)

    def visit_Name(self, node):
        self.names[node.id] = 1
        # ast.NodeVisitor.generic_visit(self, node)
        self.generic_visit(node)


class detectIllegal(ast.NodeVisitor):

    def __init__(self):
        self.illegal = False

    def visit_FunctionDef(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_ExceptHandler(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_ClassDef(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Return(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Delete(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Assign(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_AugAssign(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Print(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_For(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_While(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_If(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_With(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Raise(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_TryExcept(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_TryFinally(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Assert(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Import(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_ImportFrom(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Exec(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Global(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Pass(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Break(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Continue(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_BoolOp(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_BinOp(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_UnaryOp(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Lambda(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_IfExp(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Dict(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Set(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_ListComp(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_SetComp(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_DictComp(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_GeneratorExp(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Yield(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Compare(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Call(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Repr(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_List(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Tuple(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)


class detectIllegalQuery(ast.NodeVisitor):

    def __init__(self):
        self.illegal = False

    def visit_FunctionDef(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_ExceptHandler(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_ClassDef(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Return(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Delete(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Assign(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_AugAssign(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Print(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_For(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_While(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_If(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_With(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Raise(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_TryExcept(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_TryFinally(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Assert(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Import(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_ImportFrom(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Exec(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Global(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Pass(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Break(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Continue(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Lambda(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_IfExp(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Dict(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Set(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_ListComp(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_SetComp(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_DictComp(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_GeneratorExp(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Yield(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Call(self, node):
        try:
            assert node.func.attr in ('In', 'Like', 'And', 'Or')
        except:
            self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Repr(self, node):
        self.illegal = True
        ast.NodeVisitor.generic_visit(self, node)



# mako/util.py
# Copyright 2006-2022 the Mako authors and contributors <see AUTHORS file>
#
# This module is part of Mako and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php
from ast import parse
import codecs
import collections
import operator
import os
import re
import timeit

from .compat import importlib_metadata_get


def update_wrapper(decorated, fn):
    decorated.__wrapped__ = fn
    decorated.__name__ = fn.__name__
    return decorated


class PluginLoader:
    def __init__(self, group):
        self.group = group
        self.impls = {}

    def load(self, name):
        if name in self.impls:
            return self.impls[name]()

        for impl in importlib_metadata_get(self.group):
            if impl.name == name:
                self.impls[name] = impl.load
                return impl.load()

        from docassemble.base.mako import exceptions

        raise exceptions.RuntimeException(
            "Can't load plugin %s %s" % (self.group, name)
        )

    def register(self, name, modulepath, objname):
        def load():
            mod = __import__(modulepath)
            for token in modulepath.split(".")[1:]:
                mod = getattr(mod, token)
            return getattr(mod, objname)

        self.impls[name] = load


def verify_directory(dir_):
    """create and/or verify a filesystem directory."""

    tries = 0

    while not os.path.exists(dir_):
        try:
            tries += 1
            os.makedirs(dir_, 0o755)
        except:
            if tries > 5:
                raise


def to_list(x, default=None):
    if x is None:
        return default
    if not isinstance(x, (list, tuple)):
        return [x]
    else:
        return x


class memoized_property:

    """A read-only @property that is only evaluated once."""

    def __init__(self, fget, doc=None):
        self.fget = fget
        self.__doc__ = doc or fget.__doc__
        self.__name__ = fget.__name__

    def __get__(self, obj, cls):
        if obj is None:
            return self
        obj.__dict__[self.__name__] = result = self.fget(obj)
        return result


class memoized_instancemethod:

    """Decorate a method memoize its return value.

    Best applied to no-arg methods: memoization is not sensitive to
    argument values, and will always return the same value even when
    called with different arguments.

    """

    def __init__(self, fget, doc=None):
        self.fget = fget
        self.__doc__ = doc or fget.__doc__
        self.__name__ = fget.__name__

    def __get__(self, obj, cls):
        if obj is None:
            return self

        def oneshot(*args, **kw):
            result = self.fget(obj, *args, **kw)

            def memo(*a, **kw):
                return result

            memo.__name__ = self.__name__
            memo.__doc__ = self.__doc__
            obj.__dict__[self.__name__] = memo
            return result

        oneshot.__name__ = self.__name__
        oneshot.__doc__ = self.__doc__
        return oneshot


class SetLikeDict(dict):

    """a dictionary that has some setlike methods on it"""

    def union(self, other):
        """produce a 'union' of this dict and another (at the key level).

        values in the second dict take precedence over that of the first"""
        x = SetLikeDict(**self)
        x.update(other)
        return x


class FastEncodingBuffer:

    """a very rudimentary buffer that is faster than StringIO,
    and supports unicode data."""

    def __init__(self, encoding=None, errors="strict"):
        self.data = collections.deque()
        self.encoding = encoding
        self.delim = ""
        self.errors = errors
        self.write = self.data.append

    def truncate(self):
        self.data = collections.deque()
        self.write = self.data.append

    def getvalue(self):
        if self.encoding:
            return self.delim.join(self.data).encode(
                self.encoding, self.errors
            )
        else:
            return self.delim.join(self.data)


class LRUCache(dict):

    """A dictionary-like object that stores a limited number of items,
    discarding lesser used items periodically.

    this is a rewrite of LRUCache from Myghty to use a periodic timestamp-based
    paradigm so that synchronization is not really needed.  the size management
    is inexact.
    """

    class _Item:
        def __init__(self, key, value):
            self.key = key
            self.value = value
            self.timestamp = timeit.default_timer()

        def __repr__(self):
            return repr(self.value)

    def __init__(self, capacity, threshold=0.5):
        self.capacity = capacity
        self.threshold = threshold

    def __getitem__(self, key):
        item = dict.__getitem__(self, key)
        item.timestamp = timeit.default_timer()
        return item.value

    def values(self):
        return [i.value for i in dict.values(self)]

    def setdefault(self, key, value):
        if key in self:
            return self[key]
        self[key] = value
        return value

    def __setitem__(self, key, value):
        item = dict.get(self, key)
        if item is None:
            item = self._Item(key, value)
            dict.__setitem__(self, key, item)
        else:
            item.value = value
        self._manage_size()

    def _manage_size(self):
        while len(self) > self.capacity + self.capacity * self.threshold:
            bytime = sorted(
                dict.values(self),
                key=operator.attrgetter("timestamp"),
                reverse=True,
            )
            for item in bytime[self.capacity :]:
                try:
                    del self[item.key]
                except KeyError:
                    # if we couldn't find a key, most likely some other thread
                    # broke in on us. loop around and try again
                    break


# Regexp to match python magic encoding line
_PYTHON_MAGIC_COMMENT_re = re.compile(
    r"[ \t\f]* \# .* coding[=:][ \t]*([-\w.]+)", re.VERBOSE
)


def parse_encoding(fp):
    """Deduce the encoding of a Python source file (binary mode) from magic
    comment.

    It does this in the same way as the `Python interpreter`__

    .. __: http://docs.python.org/ref/encodings.html

    The ``fp`` argument should be a seekable file object in binary mode.
    """
    pos = fp.tell()
    fp.seek(0)
    try:
        line1 = fp.readline()
        has_bom = line1.startswith(codecs.BOM_UTF8)
        if has_bom:
            line1 = line1[len(codecs.BOM_UTF8) :]

        m = _PYTHON_MAGIC_COMMENT_re.match(line1.decode("ascii", "ignore"))
        if not m:
            try:
                parse(line1.decode("ascii", "ignore"))
            except (ImportError, SyntaxError):
                # Either it's a real syntax error, in which case the source
                # is not valid python source, or line2 is a continuation of
                # line1, in which case we don't want to scan line2 for a magic
                # comment.
                pass
            else:
                line2 = fp.readline()
                m = _PYTHON_MAGIC_COMMENT_re.match(
                    line2.decode("ascii", "ignore")
                )

        if has_bom:
            if m:
                raise SyntaxError(
                    "python refuses to compile code with both a UTF8"
                    " byte-order-mark and a magic encoding comment"
                )
            return "utf_8"
        elif m:
            return m.group(1)
        else:
            return None
    finally:
        fp.seek(pos)


def sorted_dict_repr(d):
    """repr() a dictionary with the keys in order.

    Used by the lexer unit test to compare parse trees based on strings.

    """
    keys = list(d.keys())
    keys.sort()
    return "{" + ", ".join("%r: %r" % (k, d[k]) for k in keys) + "}"


def restore__ast(_ast):
    """Attempt to restore the required classes to the _ast module if it
    appears to be missing them
    """
    if hasattr(_ast, "AST"):
        return
    _ast.PyCF_ONLY_AST = 2 << 9
    m = compile(
        """\
def foo(): pass
class Bar: pass
if False: pass
baz = 'mako'
1 + 2 - 3 * 4 / 5
6 // 7 % 8 << 9 >> 10
11 & 12 ^ 13 | 14
15 and 16 or 17
-baz + (not +18) - ~17
baz and 'foo' or 'bar'
(mako is baz == baz) is not baz != mako
mako > baz < mako >= baz <= mako
mako in baz not in mako""",
        "<unknown>",
        "exec",
        _ast.PyCF_ONLY_AST,
    )
    _ast.Module = type(m)

    for cls in _ast.Module.__mro__:
        if cls.__name__ == "mod":
            _ast.mod = cls
        elif cls.__name__ == "AST":
            _ast.AST = cls

    _ast.FunctionDef = type(m.body[0])
    _ast.ClassDef = type(m.body[1])
    _ast.If = type(m.body[2])

    _ast.Name = type(m.body[3].targets[0])
    _ast.Store = type(m.body[3].targets[0].ctx)
    _ast.Str = type(m.body[3].value)

    _ast.Sub = type(m.body[4].value.op)
    _ast.Add = type(m.body[4].value.left.op)
    _ast.Div = type(m.body[4].value.right.op)
    _ast.Mult = type(m.body[4].value.right.left.op)

    _ast.RShift = type(m.body[5].value.op)
    _ast.LShift = type(m.body[5].value.left.op)
    _ast.Mod = type(m.body[5].value.left.left.op)
    _ast.FloorDiv = type(m.body[5].value.left.left.left.op)

    _ast.BitOr = type(m.body[6].value.op)
    _ast.BitXor = type(m.body[6].value.left.op)
    _ast.BitAnd = type(m.body[6].value.left.left.op)

    _ast.Or = type(m.body[7].value.op)
    _ast.And = type(m.body[7].value.values[0].op)

    _ast.Invert = type(m.body[8].value.right.op)
    _ast.Not = type(m.body[8].value.left.right.op)
    _ast.UAdd = type(m.body[8].value.left.right.operand.op)
    _ast.USub = type(m.body[8].value.left.left.op)

    _ast.Or = type(m.body[9].value.op)
    _ast.And = type(m.body[9].value.values[0].op)

    _ast.IsNot = type(m.body[10].value.ops[0])
    _ast.NotEq = type(m.body[10].value.ops[1])
    _ast.Is = type(m.body[10].value.left.ops[0])
    _ast.Eq = type(m.body[10].value.left.ops[1])

    _ast.Gt = type(m.body[11].value.ops[0])
    _ast.Lt = type(m.body[11].value.ops[1])
    _ast.GtE = type(m.body[11].value.ops[2])
    _ast.LtE = type(m.body[11].value.ops[3])

    _ast.In = type(m.body[12].value.ops[0])
    _ast.NotIn = type(m.body[12].value.ops[1])


def read_file(path, mode="rb"):
    with open(path, mode) as fp:
        return fp.read()


def read_python_file(path):
    fp = open(path, "rb")
    try:
        encoding = parse_encoding(fp)
        data = fp.read()
        if encoding:
            data = data.decode(encoding)
        return data
    finally:
        fp.close()



from itertools import chain
from docassemble.base.functions import alpha, roman, item_label, comma_and_list, get_language, set_language, get_dialect, get_voice, set_country, get_country, word, comma_list, ordinal, ordinal_number, need, nice_number, quantity_noun, possessify, verb_past, verb_present, noun_plural, noun_singular, space_to_underscore, force_ask, force_gather, period_list, name_suffix, currency, currency_symbol, indefinite_article, nodoublequote, capitalize, title_case, url_of, do_you, did_you, does_a_b, did_a_b, your, her, his, is_word, get_locale, set_locale, update_locale, process_action, url_action, get_info, set_info, get_config, prevent_going_back, qr_code, action_menu_item, from_b64_json, defined, define, value, message, response, json_response, command, single_paragraph, quote_paragraphs, location_returned, location_known, user_lat_lon, interview_url, interview_url_action, interview_url_as_qr, interview_url_action_as_qr, interview_email, get_emails, get_default_timezone, user_logged_in, interface, user_privileges, user_has_privilege, user_info, action_arguments, action_argument, background_action, background_response, background_response_action, background_error_action, us, set_live_help_status, chat_partners_available, phone_number_in_e164, phone_number_formatted, phone_number_is_valid, countries_list, country_name, write_record, read_records, delete_record, variables_as_json, all_variables, language_from_browser, device, plain, bold, italic, states_list, state_name, subdivision_type, indent, raw, fix_punctuation, set_progress, get_progress, referring_url, undefine, invalidate, dispatch, yesno, noyes, split, showif, showifdef, phone_number_part, set_parts, log, encode_name, decode_name, interview_list, interview_menu, server_capabilities, session_tags, get_chat_log, get_user_list, get_user_info, set_user_info, get_user_secret, create_user, create_session, get_session_variables, set_session_variables, get_question_data, go_back_in_session, manage_privileges, redact, forget_result_of, re_run_logic, reconsider, set_title, set_save_status, single_to_double_newlines, verbatim, add_separators, store_variables_snapshot, update_terms, set_variables, language_name, run_action_in_session, static_image  # noqa: F401 # pylint: disable=unused-import
from docassemble.base.util import LatitudeLongitude, RoleChangeTracker, Name, IndividualName, Address, City, Event, Person, Thing, Individual, ChildList, FinancialList, PeriodicFinancialList, Income, Asset, Expense, Value, PeriodicValue, OfficeList, Organization, send_email, send_sms, send_fax, map_of, last_access_time, last_access_delta, last_access_days, last_access_hours, last_access_minutes, returning_user, timezone_list, as_datetime, current_datetime, date_difference, date_interval, today, month_of, day_of, dow_of, year_of, format_date, format_datetime, format_time, DARedis, DACloudStorage, DAGoogleAPI, SimpleTextMachineLearner, MachineLearningEntry, RandomForestMachineLearner, SVMMachineLearner, ocr_file, ocr_file_in_background, read_qr, get_sms_session, initiate_sms_session, terminate_sms_session, path_and_mimetype, run_python_module, pdf_concatenate, include_docx_template, start_time, zip_file, validation_error, DAValidationError, action_button_html, url_ask, overlay_pdf, DAStore, explain, clear_explanations, logic_explanation, set_status, get_status, DAWeb, DAWebError, json, re, iso_country, assemble_docx, docx_concatenate, task_performed, task_not_yet_performed, mark_task_as_performed, times_task_performed, set_task_counter, stash_data, retrieve_stashed_data, DABreadCrumbs, DAOAuth, DAObject, DAList, DADict, DAOrderedDict, DASet, DAFile, DAFileCollection, DAFileList, DAStaticFile, DAEmail, DAEmailRecipient, DAEmailRecipientList, DATemplate, DAEmpty, DALink, selections, objects_from_file, RelationshipTree, DAContext, DA, DAGlobal, transform_json_variables  # noqa: F401 # pylint: disable=unused-import
# from docassemble.base.logger import logmessage

__all__ = [
    'alpha',
    'roman',
    'item_label',
    'interview_url',
    'Court',
    'Case',
    'Document',
    'LegalFiling',
    'Person',
    'Thing',
    'Individual',
    'IndividualName',
    'Name',
    'Address',
    'Organization',
    'City',
    'Event',
    'DAObject',
    'DAList',
    'DADict',
    'DAOrderedDict',
    'DASet',
    'PartyList',
    'ChildList',
    'FinancialList',
    'PeriodicFinancialList',
    'Income',
    'Asset',
    'LatitudeLongitude',
    'RoleChangeTracker',
    'DATemplate',
    'DAEmpty',
    'DALink',
    'Expense',
    'Value',
    'PeriodicValue',
    'DAFile',
    'DAFileCollection',
    'DAFileList',
    'DAStaticFile',
    'DAEmail',
    'DAEmailRecipient',
    'DAEmailRecipientList',
    'send_email',
    'send_sms',
    'send_fax',
    'comma_and_list',
    'get_language',
    'get_dialect',
    'get_voice',
    'set_country',
    'get_country',
    'set_language',
    'word',
    'comma_list',
    'ordinal',
    'ordinal_number',
    'need',
    'nice_number',
    'quantity_noun',
    'verb_past',
    'verb_present',
    'noun_plural',
    'noun_singular',
    'space_to_underscore',
    'force_ask',
    'force_gather',
    'period_list',
    'name_suffix',
    'currency',
    'currency_symbol',
    'indefinite_article',
    'capitalize',
    'title_case',
    'url_of',
    'get_locale',
    'set_locale',
    'update_locale',
    'process_action',
    'url_action',
    'selections',
    'get_info',
    'set_info',
    'user_lat_lon',
    'location_known',
    'location_returned',
    'get_config',
    'map_of',
    'objects_from_file',
    'prevent_going_back',
    'month_of',
    'day_of',
    'dow_of',
    'year_of',
    'format_date',
    'format_datetime',
    'format_time',
    'today',
    'qr_code',
    'interview_url_as_qr',
    'interview_url_action_as_qr',
    'action_menu_item',
    'from_b64_json',
    'define',
    'defined',
    'value',
    'message',
    'response',
    'json_response',
    'command',
    'single_paragraph',
    'quote_paragraphs',
    'interview_url_action',
    'action_arguments',
    'action_argument',
    'last_access_time',
    'last_access_delta',
    'last_access_days',
    'last_access_hours',
    'last_access_minutes',
    'returning_user',
    'timezone_list',
    'as_datetime',
    'current_datetime',
    'date_difference',
    'date_interval',
    'get_default_timezone',
    'user_logged_in',
    'interface',
    'user_privileges',
    'user_has_privilege',
    'user_info',
    'task_performed',
    'task_not_yet_performed',
    'mark_task_as_performed',
    'times_task_performed',
    'set_task_counter',
    'background_action',
    'background_response',
    'background_response_action',
    'background_error_action',
    'us',
    'DARedis',
    'DACloudStorage',
    'DAGoogleAPI',
    'SimpleTextMachineLearner',
    'set_live_help_status',
    'chat_partners_available',
    'phone_number_in_e164',
    'phone_number_formatted',
    'phone_number_is_valid',
    'countries_list',
    'country_name',
    'write_record',
    'read_records',
    'delete_record',
    'variables_as_json',
    'all_variables',
    'ocr_file',
    'ocr_file_in_background',
    'read_qr',
    'get_sms_session',
    'initiate_sms_session',
    'terminate_sms_session',
    'language_from_browser',
    'device',
    'interview_email',
    'get_emails',
    'plain',
    'bold',
    'italic',
    'path_and_mimetype',
    'states_list',
    'state_name',
    'subdivision_type',
    'indent',
    'raw',
    'fix_punctuation',
    'set_progress',
    'get_progress',
    'referring_url',
    'run_python_module',
    'undefine',
    'invalidate',
    'dispatch',
    'yesno',
    'noyes',
    'split',
    'showif',
    'showifdef',
    'phone_number_part',
    'pdf_concatenate',
    'set_parts',
    'log',
    'encode_name',
    'decode_name',
    'interview_list',
    'interview_menu',
    'server_capabilities',
    'session_tags',
    'us_districts',
    'include_docx_template',
    'get_chat_log',
    'get_user_list',
    'get_user_info',
    'set_user_info',
    'get_user_secret',
    'create_user',
    'create_session',
    'get_session_variables',
    'set_session_variables',
    'go_back_in_session',
    'manage_privileges',
    'start_time',
    'zip_file',
    'validation_error',
    'DAValidationError',
    'redact',
    'forget_result_of',
    're_run_logic',
    'reconsider',
    'action_button_html',
    'url_ask',
    'overlay_pdf',
    'get_question_data',
    'set_title',
    'set_save_status',
    'single_to_double_newlines',
    'RelationshipTree',
    'DAContext',
    'DAOAuth',
    'DAStore',
    'explain',
    'clear_explanations',
    'logic_explanation',
    'set_status',
    'get_status',
    'verbatim',
    'add_separators',
    'DAWeb',
    'DAWebError',
    'json',
    're',
    'iso_country',
    'assemble_docx',
    'docx_concatenate',
    'store_variables_snapshot',
    'stash_data',
    'retrieve_stashed_data',
    'update_terms',
    'chain',
    'DABreadCrumbs',
    'set_variables',
    'language_name',
    'DA',
    'DAGlobal',
    'run_action_in_session',
    'static_image',
    'MachineLearningEntry',
    'RandomForestMachineLearner',
    'SVMMachineLearner',
    'transform_json_variables'
]

hooks = {}


class SpecialReturnObject():
    pass

RUN_PARENTS = SpecialReturnObject()


def register_hook(object_type, hook_name, hook, target):
    key = "^".join(target)
    if key not in hooks:
        hooks[key] = {}
    if object_type not in hooks[key]:
        hooks[key][object_type] = {}
    # logmessage("Setting hook for " + key + " " + object_type + " " + hook_name)
    hooks[key][object_type][hook_name] = hook


def run_hook(object_type, the_self, hook_name, target, **kwargs):
    while len(target):
        key = "^".join(target)
        # logmessage("Looking for hook for " + key + " and " + hook_name)
        if key in hooks and object_type in hooks[key] and hook_name in hooks[key][object_type]:
            # logmessage("Found a hook for " + key + " " + object_type + " " + hook_name)
            result = hooks[key][object_type][hook_name](the_self, **kwargs)
            # logmessage("Done with hook for " + key + " and " + hook_name)
            if not isinstance(result, SpecialReturnObject):
                return result
        target.pop()
    return None


class Court(DAObject):
    """Represents a court of law."""

    def init(self, *pargs, **kwargs):
        if 'jurisdiction' not in kwargs:
            self.jurisdiction = []
        super().init(*pargs, **kwargs)

    def __str__(self):
        return str(self.name)

    def in_the_court(self, **kwargs):  # pylint: disable=unused-argument
        """Returns the top line of the first page of a pleading filed in the
        court.

        """
        result = run_hook('court', self, 'in_the_court', self.jurisdiction)
        if result is None:
            return "In the " + self.name
        return result


def _add_person_and_children_of(target, output_list):
    if target not in output_list and target.identified():
        output_list.append(target)
        if hasattr(target, 'child'):
            for child in target.child.elements:
                _add_person_and_children_of(child, output_list)


class PartyList(DAList):
    """Represents a list of parties to a case.  The default object
    type for items in the list is Individual."""
    PartyClass = Individual

    def init(self, *pargs, **kwargs):
        self.object_type = self.PartyClass
        super().init(*pargs, **kwargs)


class Case(DAObject):
    """Represents a case in court."""
    PartyListClass = PartyList
    CourtClass = Court

    def init(self, *pargs, **kwargs):
        # logmessage("Case init: running")
        self.initializeAttribute('court', self.CourtClass)
        self.initializeAttribute('defendant', self.PartyListClass)
        self.initializeAttribute('plaintiff', self.PartyListClass)
        self.firstParty = self.plaintiff
        self.secondParty = self.defendant
        self.is_solo_action = False
        self.state = None
        self.action_type = 'plaintiff defendant'
        super().init(*pargs, **kwargs)

    def __str__(self):
        return str(self.case_id)

    def set_action_type(self, the_value):
        """Initializes attributes for the different types of parties in the
        case.

        """
        # logmessage("setting action type to " + the_value)
        # logmessage("set_action_type: self instanceName is " + self.instanceName)
        if the_value == 'solo petition':
            if hasattr(self, 'plaintiff'):
                del self.plaintiff
            if hasattr(self, 'defendant'):
                del self.defendant
            if hasattr(self, 'respondent'):
                del self.respondent
            if not hasattr(self, 'petitioner'):
                self.initializeAttribute('petitioner', self.PartyListClass)
            # logmessage("setting firstParty to petitioner")
            self.firstParty = self.petitioner
            # logmessage("firstParty instanceName is " + self.firstParty.instanceName)
            self.is_solo_action = True
        elif the_value == 'in re':
            if hasattr(self, 'plaintiff'):
                del self.plaintiff
            if hasattr(self, 'defendant'):
                del self.defendant
            if hasattr(self, 'respondent'):
                del self.respondent
            if not hasattr(self, 'petitioner'):
                self.initializeAttribute('petitioner', self.PartyListClass)
            self.firstParty = self.petitioner
            self.is_solo_action = True
        elif the_value == 'petition':
            if hasattr(self, 'plaintiff'):
                del self.plaintiff
            if hasattr(self, 'defendant'):
                del self.defendant
            if not hasattr(self, 'petitioner'):
                self.initializeAttribute('petitioner', self.PartyListClass)
            if not hasattr(self, 'respondent'):
                self.initializeAttribute('respondent', self.PartyListClass)
            self.firstParty = self.petitioner
            self.secondParty = self.respondent
            self.is_solo_action = True
        elif the_value == 'plaintiff defendant':
            if not hasattr(self, 'plaintiff'):
                self.initializeAttribute('plaintiff', self.PartyListClass)
            if not hasattr(self, 'defendant'):
                self.initializeAttribute('defendant', self.PartyListClass)
            self.firstParty = self.plaintiff
            self.secondParty = self.defendant
            self.is_solo_action = False
        elif the_value == 'appellee appellant':
            if hasattr(self, 'plaintiff'):
                del self.plaintiff
            if hasattr(self, 'defendant'):
                del self.defendant
            if not hasattr(self, 'appellee'):
                self.initializeAttribute('appellee', self.PartyListClass)
            if not hasattr(self, 'appellant'):
                self.initializeAttribute('appellant', self.PartyListClass)
            self.firstParty = self.appellee
            self.secondParty = self.appellant
            self.is_solo_action = False
        self.action_type = the_value

    def case_id_in_caption(self, **kwargs):
        """Returns the text for the case ID that will appear in the case
        caption.

        """
        result = run_hook('case', self, 'case_id_in_caption', self.court.jurisdiction, **kwargs)
        if result is None:
            if hasattr(self, 'case_id'):
                return word('Case No.') + " " + self.case_id
            return word('Case No.')
        return result

    def determine_court(self, **kwargs):
        """Runs code, if any exists, to determine what court has jurisdiction
        over the case.

        """
        # logmessage("determine_court 1: firstParty is " + str(self.firstParty.instanceName))
        run_hook('case', self, 'determine_court', self.court.jurisdiction, **kwargs)
        # logmessage("determine_court 2: firstParty is " + str(self.firstParty.instanceName))

    def role_of(self, party):
        """Given a person object, it looks through the parties to the
        case and returns the name of the party to which the person belongs.
        Returns "third party" if the person is not found among the parties."""
        for partyname in self.__dict__:
            if not isinstance(getattr(self, partyname), self.PartyListClass):
                continue
            if partyname in ['firstParty', 'secondParty']:
                continue
            getattr(self, partyname)._trigger_gather()
            for indiv in getattr(self, partyname).elements:
                if indiv is party:
                    return partyname
        return 'third party'

    def all_known_people(self):
        """Returns a list of all parties and their children,
        children's children, etc.  Does not force the gathering of the
        parties."""
        output_list = []
        for partyname in self.__dict__:
            if not isinstance(getattr(self, partyname), self.PartyListClass):
                continue
            for party in getattr(self, partyname).elements:
                _add_person_and_children_of(party, output_list)
        return output_list

    def parties(self):
        """Returns a list of all parties.  Gathers the parties if
        they have not been gathered yet."""
        output_list = []
        for partyname in self.__dict__:
            if not isinstance(getattr(self, partyname), self.PartyListClass):
                continue
            getattr(self, partyname)._trigger_gather()
            for indiv in getattr(self, partyname).elements:
                if indiv not in output_list:
                    output_list.append(indiv)
        return output_list

# class Jurisdiction(DAObject):
#     """Represents a jurisdiction, e.g. of a Court.  No functionality
#     implemented yet."""
#     pass


class Document(DAObject):
    """This is a base class for different types of documents."""

    def __str__(self):
        return str(self.title)


class LegalFiling(Document):
    """Represents a document filed in court."""

    def caption(self):
        """Returns a case caption for the case, for inclusion in documents."""
        # logmessage("caption: gathering first party")
        # logmessage("caption: case action_type is " + self.case.action_type)
        # logmessage("caption: case instanceName is " + self.case.instanceName)
        # logmessage("caption: instanceName is " + self.case.firstParty.instanceName)
        self.case.firstParty.gather()
        # logmessage("caption: gathered first party")
        if not self.case.is_solo_action:
            # logmessage("caption: case is solo action")
            self.case.secondParty.gather()
        output = ""
        # logmessage("caption: going to call in_the_court")
        output += "[BOLDCENTER] " + self.case.court.in_the_court(legalfiling=self, case=self.case) + "\n\n"
        output += "[BEGIN_CAPTION]"
        if self.case.action_type == 'in re':
            output += 'In re '
        output += comma_and_list(self.case.firstParty.elements, comma_string=",[NEWLINE]", and_string=word('and'), before_and=" ", after_and='[NEWLINE]') + ",[NEWLINE]"
        if self.case.action_type != 'in re':
            output += "[TAB][TAB]" + word(self.case.firstParty.as_noun()).capitalize() + "[NEWLINE] "
        if not self.case.is_solo_action:
            output += "[SKIPLINE][TAB]" + word('v.') + " [NEWLINE][SKIPLINE] "
            output += comma_and_list(self.case.secondParty.elements, comma_string=",[NEWLINE]", and_string=word('and'), before_and=" ", after_and='[NEWLINE]') + ",[NEWLINE]"
            output += "[TAB][TAB]" + word(self.case.secondParty.as_noun()).capitalize()
        output += "[VERTICAL_LINE]"
        output += self.case.case_id_in_caption(legalfiling=self)
        output += "[END_CAPTION]\n\n"
        if self.title is not None:
            output += "[BOLDCENTER] " + self.title.upper() + "\n"
        return output


def us_districts(bankruptcy=False):
    if bankruptcy:
        return ['Northern District of Alabama', 'Middle District of Alabama', 'Southern District of Alabama', 'District of Alaska', 'District of Arizona', 'Eastern & Western Districts of Arkansas', 'Central District of California', 'Eastern District of California', 'Northern District of California', 'Southern District of California', 'District of Colorado', 'District of Connecticut', 'District of Delaware', 'District of Columbia', 'Northern District of Florida', 'Middle District of Florida', 'Southern District of Florida', 'Northern District of Georgia', 'Middle District of Georgia', 'Southern District of Georgia', 'District of Guam', 'District of Hawaii', 'District of Idaho', 'Northern District of Illinois', 'Central District of Illinois', 'Southern District of Illinois', 'Northern District of Indiana', 'Southern District of Indiana', 'Northern District of Iowa', 'Southern District of Iowa', 'District of Kansas', 'Eastern District of Kentucky', 'Western District of Kentucky', 'Eastern District of Louisiana', 'Middle District of Louisiana', 'Western District of Louisiana', 'District of Maine', 'District of Maryland', 'District of Massachusetts', 'Eastern District of Michigan', 'Western District of Michigan', 'District of Minnesota', 'Northern District of Mississippi', 'Southern District of Mississippi', 'Eastern District of Missouri', 'Western District of Missouri', 'District of Montana', 'District of Nebraska', 'District of Nevada', 'District of New Hampshire', 'District of New Jersey', 'District of New Mexico', 'Eastern District of New York', 'Northern District of New York', 'Southern District of New York', 'Western District of New York', 'Eastern District of North Carolina', 'Middle District of North Carolina', 'Western District of North Carolina', 'District of North Dakota', 'District of the Northern Mariana Islands', 'Northern District of Ohio', 'Southern District of Ohio', 'Eastern District of Oklahoma', 'Northern District of Oklahoma', 'Western District of Oklahoma', 'District of Oregon', 'Eastern District of Pennsylvania', 'Middle District of Pennsylvania', 'Western District of Pennsylvania', 'District of Puerto Rico', 'District of Rhode Island', 'District of South Carolina', 'District of South Dakota', 'Eastern District of Tennessee', 'Middle District of Tennessee', 'Western District of Tennessee', 'Eastern District of Texas', 'Northern District of Texas', 'Southern District of Texas', 'Western District of Texas', 'District of Utah', 'District of Vermont', 'District of the Virgin Islands', 'Eastern District of Virginia', 'Western District of Virginia', 'Eastern District of Washington', 'Western District of Washington', 'Northern District of West Virginia', 'Southern District of West Virginia', 'Eastern District of Wisconsin', 'Western District of Wisconsin', 'District of Wyoming']
    return ['Northern District of Alabama', 'Middle District of Alabama', 'Southern District of Alabama', 'District of Alaska', 'District of Arizona', 'Eastern District of Arkansas', 'Western District of Arkansas', 'Central District of California', 'Eastern District of California', 'Northern District of California', 'Southern District of California', 'District of Colorado', 'District of Connecticut', 'District of Delaware', 'District of Columbia', 'Northern District of Florida', 'Middle District of Florida', 'Southern District of Florida', 'Northern District of Georgia', 'Middle District of Georgia', 'Southern District of Georgia', 'District of Guam', 'District of Hawaii', 'District of Idaho', 'Northern District of Illinois', 'Central District of Illinois', 'Southern District of Illinois', 'Northern District of Indiana', 'Southern District of Indiana', 'Northern District of Iowa', 'Southern District of Iowa', 'District of Kansas', 'Eastern District of Kentucky', 'Western District of Kentucky', 'Eastern District of Louisiana', 'Middle District of Louisiana', 'Western District of Louisiana', 'District of Maine', 'District of Maryland', 'District of Massachusetts', 'Eastern District of Michigan', 'Western District of Michigan', 'District of Minnesota', 'Northern District of Mississippi', 'Southern District of Mississippi', 'Eastern District of Missouri', 'Western District of Missouri', 'District of Montana', 'District of Nebraska', 'District of Nevada', 'District of New Hampshire', 'District of New Jersey', 'District of New Mexico', 'Eastern District of New York', 'Northern District of New York', 'Southern District of New York', 'Western District of New York', 'Eastern District of North Carolina', 'Middle District of North Carolina', 'Western District of North Carolina', 'District of North Dakota', 'District of the Northern Mariana Islands', 'Northern District of Ohio', 'Southern District of Ohio', 'Eastern District of Oklahoma', 'Northern District of Oklahoma', 'Western District of Oklahoma', 'District of Oregon', 'Eastern District of Pennsylvania', 'Middle District of Pennsylvania', 'Western District of Pennsylvania', 'District of Puerto Rico', 'District of Rhode Island', 'District of South Carolina', 'District of South Dakota', 'Eastern District of Tennessee', 'Middle District of Tennessee', 'Western District of Tennessee', 'Eastern District of Texas', 'Northern District of Texas', 'Southern District of Texas', 'Western District of Texas', 'District of Utah', 'District of Vermont', 'District of the Virgin Islands', 'Eastern District of Virginia', 'Western District of Virginia', 'Eastern District of Washington', 'Western District of Washington', 'Northern District of West Virginia', 'Southern District of West Virginia', 'Eastern District of Wisconsin', 'Western District of Wisconsin', 'District of Wyoming']
